/*
 * mtx2bas.c - Decode MTX BASIC from a snap file
 *
 * Copyright (C) 1999 Paul Daniels <paul@thewiz.freeserve.co.uk>
 * Copyright (C) 2012 Paul Daniels <paul.daniels40@yahoo.co.uk>
 *
 * Takes a .mtx snap file and outputs the BASIC portion
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 11/07/11 - PD Decode Noddy pages.
 * 19/05/12 - PD General code tidy up.
 * 21/06/12 - PD Clean up Noddy decoding and provide test file.
 * 22/06/12 - PD Decode Basic variables.
 * 12/07/12 - PD Decode Assem symbols.
 * 17/10/12 - PD Add switches to control debug output.
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "mtx2dasm.h"
#include "mtx2token.h"

/* A  B  C  D  E  F  G  H  I  J  K  L  M
   c1 c2 c3 c4 c5 c6 c7 c8 c9 ca cb cc cd
   N  O  P  Q  R  S  T  U  V  W  X  Y  Z
   ce cf d0 d1 d2 d3 d4 d5 d6 d7 d8 d9 da */

typedef	unsigned	char	byte;
typedef	unsigned	short	word;
typedef	signed		char	offset;

void	help_msg ();

int	handle_filename (FILE *, word *, word);
int	handle_sysvars  (FILE *, word, word, word *, word *, word *, word *,
				word *, word *, word *);
int	handle_basic    (FILE *, word *, word, word, word, word);
int	handle_noddy    (FILE *, word *, word, word, word);
int	handle_memory   (FILE *, word *, word, word);
int	handle_basvar	(FILE *, word *, word, word, word, word, word);

int	handle_byte (word *, byte *, word *, byte *, byte *, byte *);
int	handle_word (word *, byte *, word *, word *, byte *, byte *);
int	handle_nbyt (word *, byte *, word *, word, byte *, byte *);
int	handle_data (char *, word *, byte *, word *, word, int);
char	*handle_fp (char *, int, byte *);

int	DAsm (char *, word *, word, word, byte *, word);

int	search_symbols (byte, byte *, byte *, word, word);
int	handle_pseudo (byte *, word *, byte *);

#define file_unk	0
#define file_mtx	1
#define file_run	2
#define file_sdx	3
#define file_fdx	4

int	SHOW_SYSVARS = 0;
int	SHOW_BASDEBUG = 0;
int	SHOW_NODDEBUG = 0;
int	SHOW_SYMDEBUG = 0;
int	SHOW_ARRDEBUG = 0;
int	SHOW_VARDEBUG = 0;
int	SET_BASIC_LEN = 0;

main	(int argc, char **argv)

{

int	argcnt, optcnt, prmcnt;
FILE	*file;
int	filesize, filecurr;
int	filetype = file_mtx;

word	stklim, basbot, bastop;
word	nbtop, varnam, valbot, calcbot;
word	arrtop;
word	systop;
word	addr = 0;

word	code_len;

/******************************************************************************
 * Handle command line switches
 ******************************************************************************/

if (argc < 2) {
	help_msg ();
  return (0);
}

systop = 0xfb4b;

argcnt = 1;
prmcnt = 1;
if (argv[argcnt][0] == '-') {
  prmcnt++;
  optcnt = 1;
  while (argv[argcnt][optcnt]) {
    switch (argv[argcnt][optcnt]) {
      case 's':
	sscanf (argv[prmcnt], "%x", &systop);
	prmcnt++;
	optcnt++;
	break;
      case 'm':
	filetype = file_mtx;
	break;
      case 'r':
	filetype = file_run;
	break;
      case 'x':
	filetype = file_sdx;
	break;
      case 'f':
	filetype = file_fdx;
	break;
      case 'S':
	SHOW_SYSVARS = 1;
	break;
      case 'B':
	SHOW_BASDEBUG = 1;
	break;
      case 'L':
	SHOW_SYMDEBUG = 1;
	break;
      case 'N':
	SHOW_NODDEBUG = 1;
	break;
      case 'A':
	SHOW_ARRDEBUG = 1;
	break;
      case 'V':
	SHOW_VARDEBUG = 1;
	break;
      case 'a':
        SET_BASIC_LEN = 1;
	break;
      case 'h':
	help_msg ();
	return (0);
	break;
      default:
	printf ("Unknown option - '%c'\n", argv[argcnt][optcnt]);
	help_msg ();
	return (0);
	break;
    }
    optcnt++;
  }
}

if (!(file = fopen (argv[prmcnt], "rb"))) {
  printf ("mtx2bas: Could not open file - '%s'.\n", argv[prmcnt]);
  return (-1);
}

fseek (file, 0, SEEK_END);	/* seek to end of file */
filesize = ftell (file);	/* get current file pointer */
fseek (file, 0, SEEK_SET);	/* seek back to beginning of file */

printf ("mtx2bas v0.4. Options: ");
if (argv[argcnt][0] == '-') printf ("%s", argv[argcnt]);
printf (". File %s.\n\n", argv[prmcnt]);

if (filetype == file_mtx) {

/******************************************************************************
 * Handle .mtx Image
 ******************************************************************************/

  handle_filename (file, &stklim, systop);
  handle_sysvars (file, stklim, systop,
	      &basbot, &bastop, &nbtop, &varnam, &valbot, &calcbot, &arrtop);
  handle_basic (file, &addr, stklim, systop, basbot, (SET_BASIC_LEN ? arrtop : bastop));
  handle_noddy (file, &addr, stklim, systop, nbtop);
  handle_basvar (file, &addr, varnam, valbot, calcbot, arrtop, nbtop);

  filecurr = ftell (file);		/* get current file pointer */
  addr = 0x0000;
  handle_memory (file, &addr, filecurr, filesize);

} else if (filetype == file_run) {

/******************************************************************************
 * Handle .run Image
 ******************************************************************************/

  /* Read code_base */
  addr = (fgetc (file) + fgetc (file) * 256);
  /* Read code_length */
  code_len = (fgetc (file) + fgetc (file) * 256);
  printf ("Code Base: 0x%04X. Code Length: 0x%04X.\n\n", addr, code_len);
  handle_memory (file, &addr, 0, code_len);

} else if (filetype == file_sdx) {

/******************************************************************************
 * Handle sdx .bas Image
 ******************************************************************************/

  /* Read STKLIM */
  stklim = (fgetc (file) + fgetc (file) * 256);
  /* Read sysvars length */
  systop = stklim + ((fgetc (file) + fgetc (file) * 256));
  handle_sysvars (file, stklim, systop,
	    &basbot, &bastop, &nbtop, &varnam, &valbot, &calcbot, &arrtop);
  handle_basic (file, &addr, stklim, systop, basbot, (SET_BASIC_LEN ? arrtop : bastop));
  handle_noddy (file, &addr, stklim, systop, nbtop);
  handle_basvar (file, &addr, varnam, valbot, calcbot, arrtop, nbtop);

  /* printf ("BASIC Variables:\n");
  addr = varnam;
  handle_memory (file, &addr, varnam, calcbot); */

  filecurr = ftell (file);		/* get current file pointer */
  addr = 0x0000;
  handle_memory (file, &addr, filecurr, filesize);

} else if (filetype == file_fdx) {

/******************************************************************************
 * Handle fdx .bas Image
 ******************************************************************************/

} else {
  printf ("mtx2bas: Unknown file type.\n");
}

/******************************************************************************
 * Clean up
 ******************************************************************************/

fclose (file);

}

/******************************************************************************
 * Usage / Help message
 ******************************************************************************/

void	help_msg ()

{

printf ("Usage: mtx2bas [-mrxf] [-s <systop addr override>] <file>\n");
printf ("-s: Override SYSTOP, usually 0xfb4b. Determines last byte of sysvars.\n");
printf ("-m: file is a type .mtx.      -r: file is a type .run.\n");
printf ("-x: file is a type .sdx.      -f: file is a type .fdx.\n");
printf ("-S: Show System Variables.    -B: Show Basic Debugging.\n");
printf ("-L: Show Symbol Debugging.    -A: Show Array Debugging.\n");
printf ("-V: Show Variables Debugging. -N: Show Noddy Debugging.\n");
printf ("-a: basic len using ARRTOP.\n");
printf ("\nBy Paul Daniels. Version 0.4\n");

}

/******************************************************************************
 * Handle filename block
 ******************************************************************************/

int	handle_filename (FILE *file, word *stklim, word systop)

{

int	loop;
byte	filebyte;

/* Discard initial 0xff */
filebyte = fgetc (file);
if (filebyte != 0xff) {
  printf ("Warning: Snapshot doesn't not start with a 0xff (0x%2x).\n",
		 		filebyte);
}

/* Read filename */
printf ("MTX Filename: ");

for (loop = 0; loop < 15; loop++) {
  filebyte = fgetc (file);

  if (isprint (filebyte)) putchar (filebyte);
  else printf (".");
}

/* Read STKLIM */
*stklim = (fgetc (file) + fgetc (file) * 256);
printf (" STKLIM: 0x%04X. SYSTOP: 0x%04X.\n\n", *stklim, systop);

fflush (stdout);
return (0);

}

/******************************************************************************
 * Handle SYSVAR block
 ******************************************************************************/

int	handle_sysvars (FILE *file, word stklim, word systop, word *basbot,
	 									word	*bastop, word *nbtop, word *varnam, word *valbot,
									 	word	*calcbot, word *arrtop)

{

byte	*sv_data;
word	sv_len, sv_ptr, sv_addr;

byte	sv_lstpg;
word	sv_varnam, sv_valbot, sv_calcbot, sv_calcst, sv_kbdbuf;
byte	sv_iopl;
byte	sv_realby;
byte	sv_kbflag;
word	sv_stklim, sv_systop, sv_sstack;
word	sv_nbtop, sv_bastop, sv_basbot, sv_bastpo, sv_arrtop;
byte	sv_nbtpg, sv_bastpg, sv_arrtpg;
word	sv_baselin;
byte	sv_baslnp;
byte	sv_page, sv_crntpg, sv_datapg;
word	sv_goptr, sv_gosnum;
byte	sv_ctyslt;
word 	sv_pgtop, sv_dataad, sv_desave;
byte 	sv_pgn1, sv_pgn2;

if (!(sv_data = calloc (1, sv_len = systop - stklim))) {
  printf ("Could not alloc memory for sysvars (%d).\n", sv_len);
  return (-1);
}

fread (sv_data, sv_len, 1, file);

/* Need to check enough bytes have been read */

sv_addr = stklim;
sv_ptr = 0;

handle_data ("Data Before SYSVARS:", &sv_addr, sv_data, &sv_ptr, 0xfa52 - stklim, SHOW_SYSVARS);

handle_data ("CTRBADR:", &sv_addr, sv_data, &sv_ptr, 40, SHOW_SYSVARS);

handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_lstpg,   "LSTPG:   ", "Number of 32kb pages -1");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_varnam,  "VARNAM:  ", "Bottom of variable names (Normally 0xC000)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_valbot,  "VALBOT:  ", "Bottom of variable values (Normally 0xC001)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_calcbot, "CALCBOT: ", "Bottom of calculator stack (Normally 0xC001)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_calcst,  "CALCST:  ", "Top of calculator stack (Normally 0xC001)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_kbdbuf,  "KBDBUF:  ", "Addr of keyboard buffer (Normally 0xFB4B)");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 4,           "USYNT:   ", "USER cmd word syntax bytes");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 3,           "USER:    ", "USER cmd word jump sv_address");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 3,           "0xFA8C:  ", "Unused");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_iopl,    "IOPL:    ", "List Device");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_realby,  "REALBY:  ", "PANEL breakpoint byte");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_kbflag,  "KBFLAG:  ", "Key Lock statuses\n                                ; (2:Numlock,5:Scroll Lock,7:capslock)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_stklim,  "STKLIM:  ", "Top of free space (Normally 0xF8F2)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_systop,  "SYSTOP:  ", "Top of sysvars to save (Normally 0xFB4B)");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_sstack,  "SSTACK:  ", "Addr of machine stack (Z80 stack pointer)");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 3,           "USERINT: ", "Jump sv_addr used by user interrupt routine\n                                ; (See INTFFF (0xFD5E))");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 3,           "NODLOC:  ", "Jump sv_addr used for NODE expansion");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 3,           "FEXPAND: ", "Jump sv_addr used for PANEL expansion");
handle_nbyt (&sv_addr, sv_data, &sv_ptr, 3,           "USERNOD  ", "Jump sv_addr used for Noddy expansion");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_nbtop,   "NBTOP:   ", "Top of Noddy");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_nbtpg,   "IOPL:    ", "Top of Noddy page");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_bastop,  "BASTOP:  ", "Top of current Basic");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_bastpg,  "BASTPG:  ", "Top of current Basic page");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_basbot,  "BASBOT:  ", "Bottom of Basic");

handle_data ("\nBASTOP:", &sv_addr, sv_data, &sv_ptr, 32, SHOW_SYSVARS);

handle_word (&sv_addr, sv_data, &sv_ptr, &sv_arrtop,  "ARRTOP:  ", "Top of arrays");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_arrtpg,  "ARRTPG:  ", "Top of arrays page");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_baselin, "BASELIN: ", "Current line being executed");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_baslnp,  "BASLNP:  ", "Page Current line resides in");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_page,    "PAGE:    ", "Current page configuration");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_crntpg,  "CRNTPG:  ", "Current Basic page");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_pgn1,    "PGN1:    ", "Temp variable used by BASIC");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_pgn2,    "PGN2:    ", "Temp variable used by BASIC");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_pgtop,   "PGTOP:   ", "Address of top of current page");

handle_data ("\nGOSTACK:", &sv_addr, sv_data, &sv_ptr, 105, SHOW_SYSVARS);

handle_word (&sv_addr, sv_data, &sv_ptr, &sv_goptr,   "GOPTR:   ", "Addr to return to after GOSUB");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_gosnum,  "GOSNUM:  ", "No of nexted GOSUBs");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_ctyslt,  "CTYLST:  ", "Keyboard Configuration");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_dataad,  "DATAAD:  ", "Data pointer");
handle_byte (&sv_addr, sv_data, &sv_ptr, &sv_datapg,  "DATAPG:  ", "Data pointer page");
handle_word (&sv_addr, sv_data, &sv_ptr, &sv_desave,  "DESAVE:  ", "Addr in BASIC when saved");

handle_data ("\nData After SYSVARS:", &sv_addr, sv_data, &sv_ptr, sv_systop - 0xfb4b, SHOW_SYSVARS);

/* Variables required by other blocks */

*basbot = sv_basbot;
*bastop = sv_bastop;
*nbtop = sv_nbtop;
*varnam = sv_varnam;
*valbot = sv_valbot;
*calcbot = sv_calcbot;
*arrtop = sv_arrtop;

fflush (stdout);
return (0);

}

/******************************************************************************
 * Handle sysvar byte
 ******************************************************************************/
 
int	handle_byte (word *addr, byte *sv_data, word *sv_ptr, byte *sv_byte,
    			byte *sv_name, byte *sv_desc)

{

*sv_byte = sv_data[*sv_ptr];
if (SHOW_SYSVARS) printf ("0x%04X %s%02X\t\t; %s\n", *addr, sv_name,
	       			*sv_byte, sv_desc);
(*sv_ptr)++;
(*addr)++;
return (0);

}

/******************************************************************************
 * Handle sysvar N bytes
 ******************************************************************************/
 
int	handle_nbyt (word *addr, byte *sv_data, word *sv_ptr, word sv_len,
    			byte *sv_name, byte *sv_desc)

{

int loop;

if (SHOW_SYSVARS) printf ("0x%04X %s", *addr, sv_name);

for (loop = 0; loop < sv_len; loop++) {
  if (SHOW_SYSVARS) printf ("%02X ", sv_data[*sv_ptr]);
  (*addr)++;
  (*sv_ptr)++;
}

if (SHOW_SYSVARS) {
  if (sv_len == 3) printf ("       ");
  else printf ("    ");
}

if (SHOW_SYSVARS) printf ("; %s\n", sv_desc);
return (0);

}

/******************************************************************************
 * Handle sysvar word
 ******************************************************************************/
 
int	handle_word (word *addr, byte *sv_data, word *sv_ptr, word *sv_word,
    			byte *sv_name, byte *sv_desc)

{

*sv_word = sv_data[*sv_ptr] + (sv_data[(*sv_ptr)+1] << 8);
if (SHOW_SYSVARS) printf ("0x%04X %s%04X\t\t; %s\n", *addr, sv_name,
	       			*sv_word, sv_desc);
(*sv_ptr) += 2;
(*addr) += 2;
return (0);

}

/******************************************************************************
 * Handle BASIC data
 *
 * Basic lines are encoded as <length><line no><basic><FF> 
 * Assem lines are encoded as <length><line no><C2><assem len><label data><FF>
 * 
 * Memory pointers:
 * 
 * BASBOT		Usually 0x4000 or 0x8000
 * 
 * BASTOP		Top of Basic address
 * PGTOP		Top of Basic page
 * 
 * BASTPO		Array of top of Basic inc Noddy. Possibly per page.
 * NBTOP		Top of Noddy.
 * 
 * ARRTOP		Total bytes
 * 
 * DATAAD		Should be between BASBOT and BASBOT+BASTOP.
 *			BASBOT, 0x80 if not used.
 * BASELIN		Should be between BASBOT and BASBOT+BASTOP.
 * DESAVE		Should be between BASBOT and BASBOT+BASTOP.
 *			Ptr to next basic cmd to run after save.
 *
 * String variables have 64 taken off the first letter of the name,
 * e.g. LET A  $  =  "  "
 *      9C  01 24 D4 22 22
 ******************************************************************************/

int	handle_basic (FILE *file, word *addr, word stklim, word systop,
    			word basbot, word bastop)

{

word	addr2;
word	linptr, linptr2;
int	linnum, linlen, linlen2;
int	bascod;
int	asmlen, codlen;
int	loop, symlen, symflg;
byte	*ln_data = NULL;
char	buffer[1024];

*addr = basbot;
printf ("Basic:\n\n");

while (*addr < bastop) {

  if (SHOW_BASDEBUG) printf ("[%04X]: ", *addr);

  /* Get line length */

  linlen = fgetc (file) + (fgetc (file) * 256);
  if (SHOW_BASDEBUG) printf ("{%04X}: ", linlen);
  (*addr) += 2;
  linlen -= 2;	/* Account for linlen */

  if (ln_data) free (ln_data);
  ln_data = NULL;
  if (!(ln_data = calloc (1, linlen))) {
    printf ("Could not alloc memory for BASIC line (%d).\n", linlen);
    return (-1);
  }
  fread (ln_data, linlen, 1, file);

  linptr = 0;

  /* Get line number */

  linnum = (ln_data[linptr] + ln_data[linptr + 1] * 256);
  printf ("%d ", linnum);
  (*addr) += 2;
  linlen -= 2;	/* Account for linnum */
  linptr += 2;

  /* Decode line */

  while (linlen) {
    bascod = ln_data[linptr];

    (*addr)++;					/* Account for basic code */
    linlen--;
    linptr++;

    if (bascod == 0xff) { /* Should check if this is last char of line */
      printf ("\n");
    } else if (bascod == 0xc2) {

      printf ("ASSEM \n");

      asmlen = (ln_data[linptr] + ln_data[linptr + 1] * 256);

      (*addr) += 2;				/* Account for asmlen */
      linlen -= 2;
      linptr += 2;

      if (SHOW_BASDEBUG) printf ("        {%04X}: ", asmlen);
      printf ("CODE\n\n");
      fflush (stdout);

      addr2 = (*addr);				/* Temp addr */
      linptr2 = linptr + asmlen;		/* Symbol table pointer */
      linlen2 = linlen - asmlen;		/* Symbol table length */

      for (loop = 0; loop < asmlen;) {
	codlen = DAsm (ln_data, &linptr, *addr, addr2,
	    			&ln_data[linptr2], linlen2);

	(*addr) += codlen;			/* Account for z80 instr */
	linlen -= codlen;
	loop += codlen;
      }

      printf ("\nSymbols:\n");
      search_symbols ('L', buffer, &ln_data[linptr2], linlen2, addr2);

      handle_data ("\nSymbol Table dump:", addr, ln_data, &linptr, linlen,
		     			 SHOW_SYMDEBUG);

      /* All values are calculated relative to the
       * start address of the program.
       * The label table is setup as a big string
       * with the labels stored inside the labels.
       *
       * 0. number -	<string^><1 byte ptr cnt><1 byte ptrs>
       * 		<2 byte ptr cnt><2 byte ptrs>
       * 		<3 byte ptr cnt><3 byte ptrs>
       * 2. label -	<offset><string^><1 byte ptr cnt><1 byte ptrs>
       * 		<2 byte ptr cnt><2 byte ptrs>
       * 		<3 byte ptr cnt><3 byte ptrs>
       * 3. DB -	<offset><byte cnt><len of elements><element strings^>
       * 4. DW -	<offset><byte cnt><len of elements><element strings^>
       * 5. DS -	<offset><byte cnt><len of elements><element string^>
       * 8. Comment -	<offset><length><element string><0xff>
       *
       * Pointers
       *
       * The pointers are divided into three lots of 256
       * pointers i.e. the first byte specifies
       * how many one byte pointers there are then the 
       * pointers then how many two byte pointers
       * and finally how many JR.
       */

      if (SHOW_SYMDEBUG) {
	while (linptr2 < (linlen2 + asmlen)) {
	  switch (ln_data[linptr2]) {
	  /* number -	<string^><1 byte ptr cnt><1 byte ptrs>
	   * 		<2 byte ptr cnt><2 byte ptrs>
	   * 		<3 byte ptr cnt><3 byte ptrs> */
	  case 0x00:
	    linptr2++;
	    printf ("Number  (00): String:      ");
	    while (ln_data[linptr2] < 0x80) {
	      if (isprint (ln_data[linptr2])) printf ("%c", ln_data[linptr2]);
	      else printf (".");
	      linptr2++;
	    }
	    if (isprint (ln_data[linptr2] & 0x7f))
			  printf ("%c\n", ln_data[linptr2] & 0x7f);
	    else printf (".\n");
	    linptr2++;
	    printf ("              1 byte ptrs: 0x%02x - ", ln_data[linptr2]);
	    symlen = ln_data[linptr2];
	    linptr2++;
	    for (loop = 0; loop < symlen; loop++) {
	      printf ("0x%04x (0x%04x) ", ln_data[linptr2] +
			ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	      linptr2 += 2;
	    }
	    printf ("\n              2 byte ptrs: 0x%02x - ", ln_data[linptr2]);
	    symlen = ln_data[linptr2];
	    linptr2++;
	    for (loop = 0; loop < symlen; loop++) {
	      printf ("0x%04x (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	      linptr2 += 2;
	    }
	    printf ("\n              3 byte ptrs: 0x%02x - ", ln_data[linptr2]);
	    symlen = ln_data[linptr2];
	    linptr2++;
	    for (loop = 0; loop < symlen; loop++) {
	      printf ("0x%04x (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	      linptr2 += 2;
	    }
	    printf ("\n");
	    break;
	  /* labels -	<offset><string^><1 byte ptr cnt><1 byte ptrs>
	   * 		<2 byte ptr cnt><2 byte ptrs>
	   * 		<3 byte ptr cnt><3 byte ptrs> */
	  case 0x02:
	    linptr2++;
	    printf ("Label   (02): Offset:      0x%04x (0x%04x)\n",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	    linptr2 += 2;
	    printf ("              String:      ");
	    while (ln_data[linptr2] < 0x80) {
	      if (isprint (ln_data[linptr2])) printf ("%c", ln_data[linptr2]);
	      else printf (".");
	      linptr2++;
	    }
	    if (isprint (ln_data[linptr2] & 0x7f))
	      			printf ("%c\n", ln_data[linptr2] & 0x7f);
	    else printf (".\n");
	    linptr2++;
	    printf ("              1 byte ptrs: 0x%02x - ", ln_data[linptr2]);
	    symlen = ln_data[linptr2];
	    linptr2++;
	    for (loop = 0; loop < symlen; loop++) {
	      printf ("0x%04x (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	      linptr2 += 2;
	    }
	    printf ("\n              2 byte ptrs: 0x%02x - ", ln_data[linptr2]);
	    symlen = ln_data[linptr2];
	    linptr2++;
	    for (loop = 0; loop < symlen; loop++) {
	      printf ("0x%04x (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	      linptr2 += 2;
	    }
	    printf ("\n              3 byte ptrs: 0x%02x - ", ln_data[linptr2]);
	    symlen = ln_data[linptr2];
	    linptr2++;
	    for (loop = 0; loop < symlen; loop++) {
	      printf ("0x%04x (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	      linptr2 += 2;
	    }
	    printf ("\n");
	    break;
	  /* DB - <offset><byte cnt><len of elements><element strings^> */
	  /* DW - <offset><byte cnt><len of elements><element strings^> */
	  /* DS - <offset><byte cnt><len of elements><element strings^> */
	  case 0x03:
	  case 0x04:
	  case 0x05:
	    switch (ln_data[linptr2]) {
	    case 0x03:
	      printf ("DB      (03): ");
	      break;
	    case 0x04:
	      printf ("DW      (04): ");
	      break;
	    case 0x05:
	      printf ("DS      (05): ");
	      break;
	    }
	    linptr2++;
	    printf ("Offset:      0x%04x. (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	    linptr2 += 2;
	    printf ("Size: 0x%02x. ", ln_data[linptr2]);
	    linptr2++;
	    printf ("length: 0x%02x.\n", ln_data[linptr2]);
	    printf ("              String:      ");

	    memset (buffer, 0, 1024);
	    handle_pseudo (ln_data, &linptr2, buffer);
	    printf ("%s\n", buffer);

	    break;
	  /* Comment - <offset><length><element string><0xff> */
	  case 0x08:
	    linptr2++;
	    printf ("Comment (08): Offset:      0x%04x. (0x%04x) ",
			ln_data[linptr2] + ln_data[linptr2 + 1] * 256,
			addr2 + ln_data[linptr2] + ln_data[linptr2 + 1] * 256);
	    linptr2 += 2;
	    printf ("            length: 0x%02x.\n", ln_data[linptr2]);
	    linptr2++;
	    printf ("              String:      ");
	    while (ln_data[linptr2] != 0xff) {
	      if (ln_data[linptr2] > 0x80) {
		for (loop = 0; loop < (ln_data[linptr2] & 0x7f); loop++)
					  printf (" ");
	      } else if (isprint (ln_data[linptr2]))
					  printf ("%c", ln_data[linptr2]);
	      else printf (".");
	      linptr2++;
	    }
	    printf ("\n");
	    linptr2++;
	    break;
	  /* End of Symbols marker */
	  case 0xff:
	    linlen = 0;
	    linptr2 = linlen2 + asmlen;
	    break;
	  default:
	    printf ("Unknown symbol type: %d.\n", ln_data[linptr2]);
	    fflush (stdout);
	    return (-1);
	    break;
	  }
	}
      }

      linlen = 0;

    } else if (bascod >= 0x80) {
      printf ("%s", TOKENS[bascod - 0x80]);
      /* string variable starts with a letter - 64 */
    } else if ((bascod > 0) && (bascod < 0x20)) {
      printf ("%c", bascod + 64);
    } else if ((bascod >= 0x20) && (bascod <= 0x127)) {
      printf ("%c", bascod);
    }
  }
}

fflush (stdout);
return (0);

}

/****************************************************************************
 * Handle Noddy data
 ****************************************************************************/

/* Format

Page Length	16 bit. Includes length, name, contents and terminator.
Page Name	Last char has bit 7 high
Page contents	Spaces replaced by 80 + number of spaces.
Page terminator	0xff

Note a page is 39 columns x 24 lines.

*/

int	handle_noddy (FILE *file, word *addr, word stklim, word systop,
    			word nbtop)

{

int	done, loop;
word	datlen;
byte	*nod_data = NULL;
word	nodptr;
word	paglen;
int	nodlen, nodbyt;

datlen = nbtop - *addr;

if (nod_data) free (nod_data);
nod_data = NULL;
if (!(nod_data = calloc (1, datlen))) {
  printf ("Could not alloc memory for Noddy pages (%d).\n", datlen);
  return (-1);
}
fread (nod_data, datlen, 1, file);

nodptr = 0;

printf ("Noddy:\n");

while (nodptr < datlen) {
  if (SHOW_NODDEBUG) printf ("[%04X]: ", *addr);
  paglen = (nod_data[nodptr] + nod_data[nodptr + 1] * 256);
  if (SHOW_NODDEBUG) printf ("{%4X}: ", paglen);
  (*addr) += 2;
  nodptr += 2;
  paglen -= 2;
  /* Output page name */
  done = 0;
  while (!done) {
    nodbyt = nod_data[nodptr];
    (*addr)++;
		nodptr++;
    paglen--;
    if (nodbyt > 127) {
      printf ("%c\n", nodbyt & 0x7f);
      done = 1;
    } else {
      printf ("%c", nodbyt);
    }
  }
  printf ("---------------------------------------");
  nodlen = 0;
  while (paglen > 1) {
    nodbyt = nod_data[nodptr];
    (*addr)++;
    nodptr++;
    paglen--;
    if (nodbyt > 127) {
      for (loop = nodbyt & 0x7f; loop; loop--) {
        if (!(nodlen  % 39)) printf ("\n");
        printf (" ");
        nodlen++;
      }
    } else {
      if (!(nodlen  % 39)) printf ("\n");
      printf ("%c", nodbyt);
      nodlen++;
    }
  }
  nodbyt = nod_data[nodptr]; /* Should be 0xff */
  (*addr)++;
  nodptr++;
  printf ("\n---------------------------------------\n");
}
printf ("\n");

fflush (stdout);
return (0);

}

/****************************************************************************
 * Handle basic variables
 ****************************************************************************/

int	handle_basvar (FILE *file, word *addr, word varnam, word valbot,
    			word calcbot, word arrtop, word nbtop)

{

word	addr2;
byte	*arrdata;
word	arrlen;
word	arrptr;

byte	varbyt;
byte	*vardata;
word	varlen;
word	varptr, valptr;

char	buffer[256], buffer2[256];
int	bufptr;

word	loop;
int	loop2;
word	str_len, str_addr, str_ptr;

word	ind_cnt, ind_addr, ind_total, *ind_max, *ind_curr, ind_len;

if (!(arrdata = calloc (1, arrlen = arrtop - ((*addr) & 0x3fff)))) {
  printf ("Could not alloc memory for array variables (%d).\n", arrlen);
  return (-1);
}

fread (arrdata, arrlen, 1, file);

arrptr = 0;
handle_data ("Array Variables Dump:", addr, arrdata, &arrptr, arrlen,
    				SHOW_ARRDEBUG);

if (!(vardata = calloc (1, varlen = calcbot - varnam))) {
  printf ("Could not alloc memory for basic variables (%d).\n", varlen);
  return (-1);
}

fread (vardata, varlen, 1, file);

varptr = 0;
*addr = varnam;
handle_data ("Basic Variables Dump:", addr, vardata, &varptr, varlen,
    				SHOW_VARDEBUG);
*addr = varnam;

varptr = 0;
valptr = calcbot - varnam - 1;

bufptr = 0;

printf ("Basic Variables:\n");

addr2 = *addr;
while (*addr < valbot - 1) {
  varbyt = vardata[varptr];
  varptr++;
  (*addr)++;

  if (varbyt > 0xe0) {                    /* numeric dimension */
    buffer[bufptr] = '\0';
    bufptr = 0;
    ind_cnt = vardata[valptr - 1] + (vardata[valptr] << 8);

    ind_addr = vardata[valptr - 4] + (vardata[valptr - 3] << 8) -
      					(nbtop & 0x3fff);

    ind_max = calloc (1, ind_cnt * 2);
    ind_curr = calloc (1, ind_cnt * 2);
    ind_total = 1;

    printf ("0x%04X DIM %c%s(", addr2, varbyt - 0xa0, buffer);
    addr2 = *addr + 1;

    for (loop = 0; loop < ind_cnt; loop++) {
      ind_max[loop] = arrdata[ind_addr] + (arrdata[ind_addr + 1] << 8);
      if (loop) printf (",");
      printf ("%d", ind_max[loop]);
      ind_curr[loop] = 1;
      ind_total *= ind_max[loop];
      ind_addr += 2;
    }

    printf (") (0x%02X 0x%02X 0x%02X 0x%02X 0x%02X)\n", vardata[valptr - 4],
                              vardata[valptr - 3], vardata[valptr - 2],
                              vardata[valptr - 1], vardata[valptr]);

    for (loop = 0; loop < ind_total; loop++) {
      handle_fp (buffer2, 256, &arrdata[ind_addr]);
      if (strcmp (buffer2, "0")) {
        printf ("          LET %c%s(", varbyt - 0xa0, buffer);
        for (loop2 = 0; loop2 < ind_cnt; loop2++) {
          if (loop2) printf (",");
          printf ("%d", ind_curr[loop2]);
        }
        printf (")=");
        printf ("%s\n", buffer2);
      }
      ind_addr += 5;
      for (loop2 = ind_cnt - 1; loop2 > -1; loop2--) {
        ind_curr[loop2]++;
        if (ind_curr[loop2] > ind_max[loop2]) {
          ind_curr[loop2] = 1;
        } else {
          loop2 = -1;
        }
      }
    }

    valptr -= 5;

  } else if (varbyt > 0xc0) {              /* numeric */

    buffer[bufptr] = '\0';
    bufptr = 0;
    handle_fp (buffer2, 256, &vardata[valptr - 4]);
    printf ("0x%04X LET %c%s=%s (", addr2, varbyt - 0x80, buffer, buffer2);
    addr2 = *addr + 1;

    /* handle_fp (&vardata[valptr - 4]); */

    printf ("0x%02X 0x%02X 0x%02X 0x%02X 0x%02X)\n", vardata[valptr - 4],
                              vardata[valptr - 3], vardata[valptr - 2],
                              vardata[valptr - 1], vardata[valptr]);

    valptr -= 5;

  } else if (varbyt > 0xa0) {              /* string Dim */
    buffer[bufptr] = '\0';
    bufptr = 0;
    ind_cnt = vardata[valptr - 1] + (vardata[valptr] << 8) - 1;

    ind_addr = vardata[valptr - 4] + (vardata[valptr - 3] << 8) -
      					(nbtop & 0x3fff);

    ind_max = calloc (1, ind_cnt * 2);
    ind_curr = calloc (1, ind_cnt * 2);
    ind_total = 1;

    printf ("0x%04X DIM %c%s$(", addr2, varbyt - 0x60, buffer);
    addr2 = *addr + 1;

    for (loop = 0; loop < ind_cnt; loop++) {
      ind_max[loop] = arrdata[ind_addr] + (arrdata[ind_addr + 1] << 8);
      if (loop) printf (",");
      printf ("%d", ind_max[loop]);
      ind_curr[loop] = 1;
      ind_total *= ind_max[loop];
      ind_addr += 2;
    }

    if (loop) printf (",");
    printf ("%d) (0x%02X 0x%02X 0x%02X 0x%02X 0x%02X)\n",
			    arrdata[ind_addr] + (arrdata[ind_addr + 1] << 8),
			    vardata[valptr - 4],
			    vardata[valptr - 3], vardata[valptr - 2],
			    vardata[valptr - 1], vardata[valptr]);

    ind_len = arrdata[ind_addr] + (arrdata[ind_addr + 1] << 8);
    ind_addr += 2;

    for (loop = 0; loop < ind_total; loop++) {
      if (arrdata[ind_addr]) {
        printf ("          LET %c%s$(", varbyt - 0x60, buffer);
        for (loop2 = 0; loop2 < ind_cnt; loop2++) {
          if (loop2) printf (",");
          printf ("%d", ind_curr[loop2]);
        }
        printf (")=\"");
        for (loop2 = 0; loop2 < ind_len; loop2++) {
          if (isprint (arrdata[ind_addr])) putchar (arrdata[ind_addr]);
          ind_addr++;
        }
        printf ("\"\n");
      } else {
        ind_addr += ind_len;
      }
      for (loop2 = ind_cnt - 1; loop2 > -1; loop2--) {
        ind_curr[loop2]++;
        if (ind_curr[loop2] > ind_max[loop2]) {
          ind_curr[loop2] = 1;
        } else {
          loop2 = -1;
        }
      }
    }

    valptr -= 5;

  } else if (varbyt > 0x80) {              /* string */
    buffer[bufptr] = '\0';
    bufptr = 0;
    printf ("0x%04X LET %c%s$=\"", addr2, varbyt - 0x40, buffer);
    addr2 = *addr + 1;
    str_len = vardata[valptr - 1];    /* is length 8 or 16 bit? */
    str_addr = (vardata[valptr - 3] << 8) + vardata[valptr - 4];
    str_ptr = str_addr - (nbtop & 0x3fff);

    for (loop = 0; loop < str_len; loop++)
      					putchar (arrdata[str_ptr + 2 + loop]);

    printf ("\" (0x%02X 0x%02X 0x%02X 0x%02X 0x%02X)\n", vardata[valptr - 4],
                              vardata[valptr - 3], vardata[valptr - 2],
                              vardata[valptr - 1], vardata[valptr]);

    valptr -= 5;

  } else {
    buffer[bufptr] = varbyt;
    bufptr++;
  }
}

return (0);

}

/****************************************************************************
 * Handle Floating Point
 *
 * In memory a FP is represented as M4 M3 M2 M1 E1
 * The number is normalised so that the first binary 1 of the number
 * is in the left most bit, with the exponent increased accordingly.
 * As this bit is always a 1, it is then used to determine the numbers
 * sign.
 * The exponent is 1 - 127 when the number is normalised (shifted) to
 * the right, and 128 - 255 when normalised (shifted) to the right.
 *
 *Exponent S Mantissa
 *  1 = 00 00 00 00 81 10000001 0 0000000 00000000 00000000 00000000
 * 10 = 00 00 00 20 84 10000100 0 0100000 00000000 00000000 00000000
 * -1 = 00 00 00 80 81 10000001 1 0000000 00000000 00000000 00000000
 * 50 = 00 00 00 48 86 10000110 0 1001000 00000000 00000000 00000000
 ****************************************************************************/

char	*handle_fp (char *buffer, int length, byte *fpmem)

{

unsigned int	val_exp, val_sgn, val_man, val_int, val_fra;
double		val_fra2;

memset (buffer, '\0', length);
if (!fpmem[4] && !fpmem[3] && !fpmem[2] && !fpmem[1] && !fpmem[0]) {
  buffer[0]='0';
} else {
  val_exp = fpmem[4] & 0x7f;  /* 1 - 127 normalised to right, e.g. <= 0. */
			      /* 128 - 255 - normalised to the left, e.g. > 0 */
  val_sgn = fpmem[3] & 0x80;
  val_man = ((fpmem[3] | 0x80) << 24) + (fpmem[2] << 16) +
					  (fpmem[1] << 8) + fpmem[0];

  if (fpmem[4] > 128) {
    val_int = val_man >> (32 - val_exp);
    val_fra = val_man << val_exp;
  } else {
    val_int = 0;
    val_fra = val_man >> (128 - val_exp);
  }
  val_fra2 = (double)val_fra / (double)0x100000000;

  sprintf (buffer, "%c%9.9lf", val_sgn ? '-' : ' ', (double)val_int + val_fra2);
}

return (buffer);

}

/****************************************************************************
 * Handle generic memory block
 ****************************************************************************/

int	handle_memory (FILE *file, word *addr, word addr_start,
    			word addr_finish)

{

byte	*memdata;
word	memlen, memptr;

if (!(memdata = calloc (1, memlen = addr_finish - addr_start))) {
  printf ("Could not alloc memory for memory (%d).\n", memlen);
  return (-1);
}

fread (memdata, memlen, 1, file);

memptr = 0;

handle_data ("Remaining File Dump:", addr, memdata, &memptr, memlen, 1);

printf ("\n");

fflush (stdout);
return (0);

}

/****************************************************************************
 * Handle hex dump of memory
 ****************************************************************************/

int	handle_data (char *desc, word *addr, byte *dumpdata,
    				word *dumpptr, word datalen, int showflg)

{

int	loop, loop2, count;
byte	chrbuf[17];

chrbuf[0] = '\0';
chrbuf[16] = '\0';
count = 16;

if (showflg) printf ("%s\n", desc);

for (loop = 0; loop < datalen; loop++) {
  if (!(loop % 16)) {
    for (loop2 = count; loop2 < 16; loop2++) if (showflg) printf ("   ");
    if (showflg) printf ("\t%s\n0x%04X ", chrbuf, *addr);
    for (loop2 = 0; loop2 < 16; loop2++) chrbuf[loop2] = ' ';
    count = 0;
  }
  if (showflg) printf ("%02X ", dumpdata[*dumpptr]);
  count++;
  if (isprint (dumpdata[*dumpptr])) chrbuf[loop % 16] = dumpdata[*dumpptr];
  else chrbuf[loop % 16] = '.';
  (*addr)++;
  (*dumpptr)++;
}
if (loop) {
  for (loop2 = count; loop2 < 16; loop2++) if (showflg) printf ("   ");
  if (showflg) printf ("\t%s\n\n", chrbuf);
} else {
  if (showflg) printf ("\n");
}

fflush (stdout);
return (0);

}

/******************************************************************************
 * DAsm() will disassemble the code at address A. It will return the number
 * of bytes disassembled.
 ******************************************************************************/

int	DAsm (char *ln_data, word *linptr, word address,
    		word offaddr, byte *symbols, word symlen)

{

char	tmp_mnem[128],hexbuf[257],index_reg,*init_mnem,*code_mnem;
byte	index_flg, Offset, inbyt, inbyt2;
word	upd_addr;
char	byte_str[32];
char	tmparea[5];
char	code_str[257];
char	comment[257], label[257];
word	symptr, ptrcnt, offset;
word	addr2, comptr, labptr, pseudo_cnt;
int	loop, loop2, symtyp;

upd_addr = address;
index_reg = '\0';
index_flg = 0;
byte_str[0] = '\0';

/* Walk through the symbols table looking for a type 0 and 2, 3 or 4 or 5
 * and 8. */

symptr = 0;
comment[0] = '\0';
label[0] = '\0';
addr2 = address;

pseudo_cnt = 0;

search_symbols ('c', comment, symbols, symlen, address - offaddr);
search_symbols ('l', label, symbols, symlen, address - offaddr);
pseudo_cnt = search_symbols ('d', code_str, symbols, symlen, address - offaddr);

/* Handle Pseudo op */

/* If pseudo op flag set (no of bytes)
 * Construct return string and return. Note it could be multiple lines
 * as only 4 bytes displayed per line
 */

if (pseudo_cnt) {
  byte_str[0] = '\0';
  for (loop = 0; loop < pseudo_cnt; loop++) {
    inbyt = ln_data[(*linptr)++];
    sprintf (tmparea, "%02x ", inbyt);
    strcat (byte_str, tmparea);
    if ((loop % 4) == 3) {
      printf ("0x%04X  %s", upd_addr, byte_str);
      if (loop < 4) {
	printf ("%-8.8s %-20.20s %-30.30s", label, code_str, comment);
      }
      printf ("\n");
      upd_addr += 4;
      byte_str[0] = '\0';
    }
  }
  if ((loop % 4) != 0) {
    printf ("0x%04X  %s", upd_addr, byte_str);
    upd_addr += (loop % 4);
    if (loop < 4) {
      for (loop = (loop - 1) % 4; loop < 3; loop++) printf ("   ");
      printf ("%-8.8s %-20.20s %-30.30s", label, code_str, comment);
    }
    printf ("\n");
  }
  return (upd_addr-address);
}

/* Lookup Mnemonic */

printf ("0x%04X  ", address);

inbyt = ln_data[(*linptr)++];
switch (inbyt) {
case 0xCB:
  strcat (byte_str, "CB ");
  upd_addr++;
	inbyt = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);
  init_mnem= MnemonicsCB[inbyt];
  upd_addr++;
  break;
case 0xED:
  strcat (byte_str, "ED ");
  upd_addr++;
	inbyt = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);
  init_mnem= MnemonicsED[inbyt];
  upd_addr++;
  break;
case 0xDD:
  strcat (byte_str, "DD ");
  upd_addr++;
  index_reg = 'X';
	inbyt2 = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt2);
  strcat (byte_str, tmparea);
  if ((inbyt = inbyt2) != 0xCB) {
    init_mnem = MnemonicsXX[inbyt];
    upd_addr++;
  } else {
    upd_addr++;
    Offset = inbyt;
    index_flg = 1;
		inbyt2 = ln_data[(*linptr)++];
    sprintf (tmparea, "%02X ", inbyt2);
    strcat (byte_str, tmparea);
    init_mnem = MnemonicsXCB[inbyt2];
    upd_addr++;
  }
  break;
case 0xFD:
  strcat (byte_str, "FD ");
  upd_addr++;
  index_reg = 'Y';
	inbyt2 = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt2);
  strcat (byte_str, tmparea);
  if ((inbyt = inbyt2) != 0xCB) {
    init_mnem = MnemonicsXX[inbyt];
    upd_addr++;
  } else {
    upd_addr++;
    Offset = inbyt;
    index_flg = 1;
		inbyt2 = ln_data[(*linptr)++];
    sprintf (tmparea, "%02X ", inbyt2);
    strcat (byte_str, tmparea);
    init_mnem = MnemonicsXCB[inbyt2];
    upd_addr++;
  }
  break;
default:
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);
  upd_addr++;
  init_mnem = Mnemonics[inbyt];
}

/* Check for index byte */

if (code_mnem = strchr (init_mnem,'^')) {
  strncpy (tmp_mnem,init_mnem,code_mnem-init_mnem);
  tmp_mnem[code_mnem-init_mnem] = '\0';
	inbyt = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);
  search_symbols ('^', hexbuf, symbols, symlen, upd_addr - offaddr);
  if (!hexbuf[0]) sprintf (hexbuf, "%d", inbyt);
  upd_addr++;
  strcat (tmp_mnem, hexbuf);
  strcat (tmp_mnem,code_mnem+1);

} else strcpy (tmp_mnem, init_mnem);

/* Check for index register */

if (code_mnem = strchr (tmp_mnem, '%')) *code_mnem=index_reg;

/* Check for byte */

if (code_mnem = strchr (tmp_mnem, '*')) {
  strncpy (code_str,tmp_mnem,code_mnem-tmp_mnem);
  code_str[code_mnem-tmp_mnem] = '\0';
  inbyt = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);

  /* Need a routine to search the symbol table
   * looking for type 0 and 2 entries.
   * Will need to know the address being checked.
   */

  search_symbols ('*', hexbuf, symbols, symlen, upd_addr - offaddr);
  if (!hexbuf[0]) sprintf (hexbuf, "%d", inbyt);
  upd_addr++;
  strcat (code_str, hexbuf);
  strcat (code_str, code_mnem+1);

/* Check for relative jump */

} else if (code_mnem = strchr (tmp_mnem, '@')) {
  strncpy (code_str,tmp_mnem,code_mnem-tmp_mnem);
  code_str[code_mnem-tmp_mnem] = '\0';
	inbyt = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);
  if (!index_flg) Offset = inbyt;
  search_symbols ('@', hexbuf, symbols, symlen, upd_addr - offaddr);
  if (!hexbuf[0]) {
    strcat (code_str, Offset & 0x80 ? "-" : "+");
    index_flg = Offset & 0x80 ? 256 - Offset : Offset;
    sprintf (hexbuf, "%d", index_flg);
  }
  upd_addr++;
  strcat (code_str,hexbuf);
  strcat (code_str, code_mnem+1);

/* Check for word */

} else if (code_mnem = strchr(tmp_mnem,'~')) {
  strncpy (code_str, tmp_mnem, code_mnem-tmp_mnem);
  code_str[code_mnem-tmp_mnem] = '\0';
  inbyt = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt);
  strcat (byte_str, tmparea);
  inbyt2 = ln_data[(*linptr)++];
  sprintf (tmparea, "%02X ", inbyt2);
  strcat (byte_str, tmparea);
  search_symbols ('~', hexbuf, symbols, symlen, upd_addr - offaddr);
  if (!hexbuf[0]) sprintf (hexbuf, "%d", inbyt + inbyt2 * 256);
  strcat (code_str, hexbuf);
  strcat (code_str, code_mnem+1);
  upd_addr += 2;

} else strcpy(code_str,tmp_mnem);

printf ("%-11.11s %-8.8s %-20.20s %-30.30s\n", byte_str, label,
    		code_str, comment);

return (upd_addr-address);

}

/******************************************************************************
 * Search assem symbol table for various entries
 ******************************************************************************/

int	search_symbols (byte dasmtyp, byte *buffer, byte *symbols,
    				word symlen, word upd_addr)

{

word	symptr, savptr, symaddr, offset, comptr;
byte	symtyp, ptrcnt, loop, loop2, bytecnt;
int	labflg;

buffer[0] = '\0';
bytecnt = 0;
labflg = 3;

symptr = 0;
while (symptr < symlen) {

  symtyp = symbols[symptr];

  symptr++;
  switch (symtyp) {
  /* number - <string^><1 byte ptr cnt><1 byte ptrs><2 byte ptr cnt>
   * <2 byte ptrs><3 byte ptr cnt><3 byte ptrs> */
  case 0:
    savptr = symptr;
    while (symbols[symptr] < 0x80) symptr++;
    symptr++;
    ptrcnt = symbols[symptr++];
    switch (dasmtyp) {
      case '*':
      case '@':
      case '^':
        for (loop = 0; loop < ptrcnt; loop++) {
          symaddr = symbols[symptr] + symbols[symptr + 1] * 256;
          symptr +=2;
          if (symaddr == upd_addr) {
            loop2 = 0;
            while (symbols[savptr] < 0x80) {
              buffer[loop2++] = symbols[savptr++];
            }
            buffer[loop2++] = symbols[savptr++] & 0x7f;
            buffer[loop2++] = 0;
          }
        }
        break;
      default:
        symptr += (ptrcnt * 2);
        break;
    }
    ptrcnt = symbols[symptr++];
    switch (dasmtyp) {
      case '~':
        for (loop = 0; loop < ptrcnt; loop++) {
          symaddr = symbols[symptr] + symbols[symptr + 1] * 256;
          symptr +=2;
          if (symaddr == upd_addr) {
            loop2 = 0;
            while (symbols[savptr] < 0x80) {
              buffer[loop2++] = symbols[savptr++];
            }
            buffer[loop2++] = symbols[savptr++] & 0x7f;
            buffer[loop2++] = 0;
          }
        }
        break;
      default:
        symptr += (ptrcnt * 2);
        break;
    }
    ptrcnt = symbols[symptr++];
    switch (dasmtyp) {
      case '@':
        for (loop = 0; loop < ptrcnt; loop++) {
          symaddr = symbols[symptr] + symbols[symptr + 1] * 256;
          symptr +=2;
          if (symaddr == upd_addr) {
            loop2 = 0;
            while (symbols[savptr] < 0x80) {
              buffer[loop2++] = symbols[savptr++];
            }
            buffer[loop2++] = symbols[savptr++] & 0x7f;
            buffer[loop2++] = 0;
          }
        }
        break;
      default:
        symptr += (ptrcnt * 2);
        break;
    }
    break;
  /* label - <offset><string^><1 byte ptr cnt><1 byte ptrs><2 byte ptr cnt>
   * <2 byte ptrs><3 byte ptr cnt><3 byte ptrs> */
  case 2:
    offset = symbols[symptr] + symbols[symptr + 1] * 256;
    symptr += 2;                /* offset */
    savptr = symptr;
    while (symbols[symptr] < 0x80) symptr++;
    symptr++;
    ptrcnt = symbols[symptr++];
    switch (dasmtyp) {
      case '*':
      case '^':
        for (loop = 0; loop < ptrcnt; loop++) {
          symaddr = symbols[symptr] + symbols[symptr + 1] * 256;
          symptr +=2;
          if (symaddr == upd_addr) {
            loop2 = 0;
            while (symbols[savptr] < 0x80) {
              buffer[loop2++] = symbols[savptr++];
            }
            buffer[loop2++] = symbols[savptr++] & 0x7f;
            buffer[loop2++] = 0;
          }
        }
        break;
      case 'l':
	if (offset == upd_addr) {
	  loop2 = 0;
	  while (symbols[savptr] < 0x80) {
	    buffer[loop2++] = symbols[savptr++];
	  }
	  buffer[loop2++] = symbols[savptr] & 0x7f;
	  buffer[loop2++] = ':';
	  buffer[loop2++] = '\0';
	}
        symptr += (ptrcnt * 2);
	break;
      case 'L':
	loop2 = 0;
	while (symbols[savptr] < 0x80) {
	  printf ("%c", symbols[savptr]);
	  fflush (stdout);
	  savptr++;
	  loop2++;
	}
	printf ("%c", symbols[savptr] & 0x7f);

	for (; loop2 < 11; loop2++) printf (" ");
	printf ("  %04X  ", offset + upd_addr);
	labflg = (labflg++ % 4);
	if (labflg == 3) printf ("\n");
        symptr += (ptrcnt * 2);
	break;
      default:
        symptr += (ptrcnt * 2);
        break;
    }
    ptrcnt = symbols[symptr++];
    switch (dasmtyp) {
      case '~':
        for (loop = 0; loop < ptrcnt; loop++) {
          symaddr = symbols[symptr] + symbols[symptr + 1] * 256;
          symptr +=2;
          if (symaddr == upd_addr) {
            loop2 = 0;
            while (symbols[savptr] < 0x80) {
              buffer[loop2++] = symbols[savptr++];
            }
            buffer[loop2++] = symbols[savptr++] & 0x7f;
            buffer[loop2++] = 0;
          }
        }
        break;
      default:
        symptr += (ptrcnt * 2);
        break;
    }
    ptrcnt = symbols[symptr++];
    switch (dasmtyp) {
      case '@':
        for (loop = 0; loop < ptrcnt; loop++) {
          symaddr = symbols[symptr] + symbols[symptr + 1] * 256;
          symptr +=2;
          if (symaddr == upd_addr) {
            loop2 = 0;
            while (symbols[savptr] < 0x80) {
              buffer[loop2++] = symbols[savptr++];
            }
            buffer[loop2++] = symbols[savptr++] & 0x7f;
            buffer[loop2++] = 0;
          }
        }
        break;
      default:
        symptr += (ptrcnt * 2);
        break;
    }
    break;
  /* DB - <offset><byte cnt><len of elements><element strings^> */
  /* DW - <offset><byte cnt><len of elements><element strings^> */
  /* DS - <offset><byte cnt><len of elements><element strings^> */
  case 3: /* DB */
  case 4: /* DW */
  case 5: /* DS */
    offset = symbols[symptr] + symbols[symptr + 1] * 256;
    symptr += 2;
    switch (dasmtyp) {
      case 'd':
	if (offset == upd_addr) {
	  buffer[0] = 'D';
	  if (symtyp == 3) buffer[1] = 'B';
	  else if (symtyp == 4) buffer[1] = 'W';
	  else buffer[1] = 'S';
	  buffer[2] = ' ';

	  bytecnt = symbols[symptr++];
	  handle_pseudo (symbols, &symptr, &buffer[3]);
	} else {
	  symptr += 1;			/* skip byte cnt */
	  ptrcnt = symbols[symptr];	/* len of elements */
	  symptr += 1 + ptrcnt;		/* Skip len & element strings */
	}
	break;
      default:
	symptr += 1;			/* skip byte cnt */
	ptrcnt = symbols[symptr];	/* len of elements */
	symptr += 1 + ptrcnt;		/* Skip len & element strings */
	break;
    }
    break;
  /* Comment - <offset><length><element string><0xff> */
  case 8:
    offset = symbols[symptr] + symbols[symptr + 1] * 256;
    symptr += 2;
    if ((dasmtyp == 'c') && (offset == upd_addr)) {
      buffer[0] = ' ';
      buffer[1] = ';';
      comptr = 2;
      for (loop = 1; loop < symbols[symptr]; loop++) {
	if (symbols[symptr + loop] < 0x80)
			      buffer[comptr++] = symbols[symptr + loop];
	else {
	  for (loop2 = 0; loop2 < (symbols[symptr + loop] & 0x7f); loop2++)
			      buffer[comptr++] = ' ';
	}
      }
      symptr++;
      buffer[comptr++] = '\0';
    } else {
      while (symbols[symptr] != 0xff) symptr++;
      symptr++;
    }
    break;
  }
}

if (labflg != 3) printf ("\n");

return (bytecnt);

}

/******************************************************************************
 * Handle pseudo assem mnemonic parameters
 ******************************************************************************/

int	handle_pseudo (byte *pseudo_data, word *pseudo_ptr, byte *buffer)

{

byte	pseudo_len, loop;
int	pseudo_flg, buf_ptr;

buf_ptr = 0;
pseudo_len = pseudo_data[*pseudo_ptr];
(*pseudo_ptr)++;
pseudo_flg = 0;
for (loop = 0; loop < pseudo_len; loop++) {
  if (pseudo_flg) {
    buffer[buf_ptr++] = ',';
    pseudo_flg = 0;
  }
  if (pseudo_data[*pseudo_ptr] & 0x80) pseudo_flg = 1;

  if (isprint (pseudo_data[*pseudo_ptr] & 0x7f))
    		buffer[buf_ptr++] = pseudo_data[*pseudo_ptr] & 0x7f;
  else buffer[buf_ptr++] = '.';

  (*pseudo_ptr)++;
}

buffer[buf_ptr] = '\0';

return (0);

}
